﻿# -*- coding: utf-8 -*-
"""
kpga.pyプログラム
ナップサック問題のＧＡによる求解プログラム 
ＧＡによって、ナップサック問題の最良解を探索します
使い方　c:\>python kpga.py < data.txt
"""
# モジュールのインポート
import math
import random
import copy

# グローバル変数
MAXVALUE = 100                  #重さと価値の最大値
N = 30                          #荷物の個数
WEIGHTLIMIT = N * MAXVALUE / 4  #重量制限
POOLSIZE = 30                   #プールサイズ  
LASTG = 50                      #打ち切り世代
MRATE = 0.01                    #突然変異の確率
SEED = 32767                    #乱数のシード
parcel = [[0 for i in range(2)] 
         for j in range(N)]     #荷物

# 下請け関数の定義
# rndn()関数
def rndn(n):
    """n未満の乱数の生成"""
    return random.randint(0,n-1)  
# rndn()関数の終わり

# initparcel()関数
def initparcel():
    """荷物の初期化"""
    i = 0
    while i < N:
        try:
            line = input()
        except EOFError:
            break #入力の終わり
        parcel[i] = [int(num) for num in line.split()]
        i += 1
    return 
# initparcel()関数の終わり

# mating()関数
def mating(pool,ngpool):
    """交叉"""
    roulette = [0 for i in range(POOLSIZE)] #ルーレット
    #ルーレットの作成
    totalfitness = 0                        #適応度の合計値
    for i in range(POOLSIZE):
        roulette[i] = evalfit(pool[i])
        #適応度の合計値を計算
        totalfitness += roulette[i]
    #選択と交叉を繰り返す
    for i in range(POOLSIZE):
        while True:   #重複の削除
            mama = selectp(roulette,totalfitness)
            papa = selectp(roulette,totalfitness)
            if mama != papa:
                break #重複なし
        #特定の２遺伝子の交叉
        crossing(pool[mama],pool[papa]
                 ,ngpool[i * 2],ngpool[i * 2 + 1])
 
    return 
# mating()関数の終わり

# evalfit()関数
def evalfit(g):
    """適応度の計算"""
    value = 0         #評価値
    weight = 0        #重量
    #各遺伝子座を調べて重量と評価値を計算
    for pos in range(N):
        weight += parcel[pos][0] * g[pos]
        value  += parcel[pos][1] * g[pos]
    #致死遺伝子の処理
    if weight >= WEIGHTLIMIT:
        value = 0
    return value
# evalfit()関数の終わり

# selectp()関数
def selectp(roulette,totalfitness):
    """親の選択"""
    acc = 0
    ball = rndn(totalfitness)
    for i in range(POOLSIZE):
        acc += roulette[i]  #適応度を積算
        if acc > ball:
            break           #対応する遺伝子   
    return i
# selectp()関数の終わり

# crossing()関数
def crossing(m,p,c1,c2):
    """特定の２染色体の交叉"""
    #交叉点の決定
    cp =rndn(N) 
    #前半部分のコピー
    for j in range(cp):
        c1[j] = m[j]
        c2[j] = p[j]
    #後半部分のコピー
    for j in range(cp,N):
        c2[j] = m[j]
        c1[j] = p[j]
    return
# crossing()関数の終わり

# mutation()関数
def mutation(ngpool):
    """突然変異"""
    for i in range(POOLSIZE*2):
        for j in range(N):
            if random.random() < MRATE:
                #反転の突然変異
                ngpool[i][j] = 1 if ngpool[i][j] == 0 else 0
    return 
# mutation()関数の終わり

# selectng()関数
def selectng(ngpool,pool):
    """次世代の選択"""
    totalfitness = 0                            #適応度の合計値
    roulette = [0 for i in range(POOLSIZE * 2)] #ルーレット
    acc = 0                                     #適応度の積算値

    #選択を繰り返す
    for i in range(POOLSIZE):
        #ルーレットの作成
        totalfitness=0
        for c in range(POOLSIZE * 2):
            roulette[c] = evalfit(ngpool[c])
            #適応度の合計値を計算
            totalfitness += roulette[c]
        #染色体を一つ選ぶ
        ball = rndn(totalfitness)
        acc = 0
        for c in range(POOLSIZE*2):
            acc += roulette[c]  #適応度を積算
            if acc > ball:
                break           #対応する遺伝子
        # 染色体のコピー
        pool[i] = copy.deepcopy(ngpool[c])
    return 
# selectng()関数の終わり

# printp()関数
def printp(pool):
    """結果出力"""
    totalfitness = 0  #適応度の合計値
    bestfit = 0       #エリート遺伝子の処理用変数
    #染色体プールの出力
    for i in range(POOLSIZE):
        fitness = evalfit(pool[i])
        if fitness > bestfit: #エリート解
            bestfit = fitness
            elite = i
        print(pool[i],"  fitness =",fitness)
        totalfitness += fitness
    #エリート解の適応度と平均適応度を出力
    print(elite,bestfit,totalfitness/POOLSIZE)
    return 
# printp()関数の終わり

# メイン実行部
#乱数の初期化
random.seed(SEED)

#変数の準備
pool = [[rndn(2) for i in range(N)] 
         for j in range(POOLSIZE)]      #染色体プール
ngpool = [[0 for i in range(N)] 
          for j in range(POOLSIZE * 2)] #次世代染色体プール
generation = 0                          #現在の世代数

#荷物の初期化
initparcel()

#打ち切り世代まで繰り返し
for generation in range(LASTG):
    print(generation,"世代")
    mating(pool,ngpool)   #交叉
    mutation(ngpool)      #突然変異
    selectng(ngpool,pool) #次世代の選択
    printp(pool)          #結果出力  
# kpga.pyの終わり
